/*
 * Copyright 2000 by Docuverse
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Docuverse ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you
 * entered into with Docuverse.
 */

package com.docuverse.min;

//import java.io.CharConversionException;
import java.io.InputStream;
import java.io.IOException;
import java.io.Reader;

/**
Custom UTF-8 input stream reader optimized for speed and size.
Methods not used by com.docuverse.min parser are not supported.
*/
final class UTF8 extends Reader
{
	private static final int	END_OF_DATA = 0;

	private InputStream		in;
	private boolean			closable;
	private byte[]			byteBuf;
	private int				bytePos;
	private int				readLen;
	private int				charEx;

	UTF8(InputStream in, boolean closable, byte[] buffer, int offset, int length)
	{
		super(in);

		this.in = in;
		this.closable = closable;
		this.byteBuf = buffer;
		this.bytePos = offset;
		this.readLen = buffer.length - (buffer.length % 1024);
		this.charEx = 0;

		buffer[offset + length] = END_OF_DATA;
	}

// Reader methods

		public final void
	close()
		throws IOException
	{
		if (in != null)
		{
			if (closable)
				in.close();
			in = null;
		}
	}

		public final int
	read(char[] cb, int offset, int length)
		throws IOException
	{
		if (in == null)
			return -1;

		int		bi = bytePos;
		byte[]	bb = byteBuf;
		int		ce = offset + length;
		int		c;
		int		n;

			// emit surrogate leftover
		if (charEx != 0)
		{
			cb[offset++] = (char)charEx;
			charEx = 0;
		}

		while (offset < ce)
		{
			if ((c = bb[bi++] & 0xff) == END_OF_DATA)
			{
				bi = 0;
				n = in.read(bb, 0, readLen);
				if (n > 0)
				{
					bb[n] = END_OF_DATA;
					continue;
				}
				else
				{
					bb[0] = END_OF_DATA;
					close();
					break;
				}
			}
				// ASCII range (most common case)
			else if (c < 128)
			{
				cb[offset++] = (char)c;
				continue;
			}
				// multibyte character
			else
			{
					// position to restart from on index exception
				int ix = bi - 1;
				try
				{
					// array boundary check via exception

						// 2 bytes?
					if ((c & 0xE0) == 0xC0)
					{
						c  = (c & 0x1F) << 6;
						c +=  bb[bi++] & 0x3F;
						cb[offset++] = (char)c;
						continue;
					}
						// 3 bytes?
					else if ((c & 0xF0) == 0xE0)
					{
						c  = (c & 0x0F) << 12;
						c += (bb[bi++] & 0x3F) << 6;
						c +=  bb[bi++] & 0x3F;
						cb[offset++] = (char)c;
						continue;
					}
						// 4 bytes
					else if ((c & 0xF8) == 0xF0)
					{
						c  = (c & 0x07) << 18;
						c += (bb[bi++] & 0x3F) << 12;
						c += (bb[bi++] & 0x3F) << 6;
						c +=  bb[bi++] & 0x3F;
						if (c > 0xffff && c <= 0x10FFFF)
						{
							c -= 0x10000;
							charEx = 0xDC00 + (c & 0x03ff);
							c = 0xD800 + (c >> 10);
							cb[offset++] = (char)c;
							if (offset < ce)
							{
								cb[offset++] = (char)charEx;
								charEx = 0;
							}
							continue;
						}
						else
							throw new IOException(
								"Illegal UTF-8 encoding " +
								Integer.toHexString(c));
					}
					else
						throw new IOException(
							"Illegal UTF-8 encoding " +
							Integer.toHexString(c));
				}
				catch (ArrayIndexOutOfBoundsException e)
				{
					// not enough data to complete the character
					// move the incomplete sequence to head of buffer
					// and read in more data

					bi = 0;
					byte b;
					while ((b = bb[ix++]) != 0)
						bb[bi++] = b;

					n = in.read(bb, bi, readLen);
					bi = 0;
					if (n > 0)
					{
						bb[n] = END_OF_DATA;
						continue;
					}
					else
					{
						close();
						break;
					}
				}
			}
		}

		bytePos = bi;
		n = length - (ce - offset);
		if (n > 0)
			return n;

		return -1;
	}
}
